import os
import requests
from urllib.parse import urljoin
from swap.common.config import Settings
from swap.common.logging import logger


class UBLClient:
    def __init__(self):
        self.base_url = Settings.UBL_URL
        self.api_key = os.getenv('SWAP_API_TOKEN')

    def url(self, path):
        return urljoin(self.base_url, path)

    def get(self, path=''):
        logger.debug(f'Getting from UBL on path: {self.url(path)}')
        params = {'api_key': self.api_key}
        joined_url = self.url(path)
        response = requests.get(joined_url, params=params)

        return response.text

    def post(self, path='', data={}):
        logger.debug(f'Posting to UBL on path: {self.url(path)}')
        params = {'api_key': self.api_key}
        joined_url = self.url(path)
        response = requests.post(joined_url, json=data, params=params)

        if response.status_code != requests.codes.ok:
            raise Exception(f'UBL API call response not OK: {response.status_code}')

        return response.text

    def get_input_datasets(self):
        return self.get('/inputdata')


class ServiceDiscoveryClient:
    def __init__(self):
        self.base_url = Settings.SERVICE_DISCOVERY_URL

    def url(self, path):
        return urljoin(self.base_url, path)

    def get(self, path=''):
        joined_url = self.url(path)
        response = requests.get(joined_url)
        return response.text

    def get_services(self):
        return self.get('/servicediscovery')


class IngestionClient:
    def __init__(self):
        self.base_url = Settings.INGESTION_URL
        self.id_token = os.getenv('SWAP_API_TOKEN')

    def url(self, path):
        return urljoin(self.base_url, path)

    def get(self, path=''):
        joined_url = self.url(path)
        response = requests.get(joined_url, headers={'id_token': self.id_token})

        return response.text

    def post(self, path='', data={}):
        response = requests.post(self.url(path), json=data, headers={'id_token': self.id_token})

        return response.text

    def post_ingestion(self, data):
        return self.post('/ingestions', data=data)

    def get_ingestion_progress(self, id):
        return self.get(f'/ingestions/{id}')



class ServiceOffersResponse:
    def __init__(self, chain_links):
        self.chain_links = chain_links

    def add_chain_link(self, chain_link):
        self.chain_links[chain_link.key] = chain_link
        setattr(self, chain_link.key, chain_link)


class ServiceOffersChainLink:
    def __init__(self, message_id, service_catalogue_uuid, service_offers):
        self.message_id = message_id
        self.service_catalogue_uuid = service_catalogue_uuid
        self.service_offers = service_offers

    def add_service_offer(self, service_offer):
        self.service_offers[service_offer.key] = service_offer


class ServiceOffer:
    def __init__(
        self, message_id, publish_time, parameters, app_message_uuid, request_message_uuid,
        originating_message_uuid, service_chain_uuid, service_provider_uuid, service_offer_uuid,
        service_request_uuid, service_provider_signature, service_offer_name,
        service_offer_description, file_uris
    ):
        self.message_id = message_id
        self.publish_time = publish_time
        self.parameters = parameters
        self.app_message_uuid = app_message_uuid
        self.request_message_uuid = request_message_uuid
        self.originating_message_uuid = originating_message_uuid
        self.service_chain_uuid = service_chain_uuid
        self.service_provider_uuid = service_provider_uuid
        self.service_offer_uuid = service_offer_uuid
        self.service_request_uuid = service_request_uuid
        self.service_provider_signature = service_provider_signature
        self.service_offer_name = service_offer_name
        self.service_offer_description = service_offer_description
        self.file_uris = file_uris


class ServiceOfferParameters:
    def __init__(self, properties):
        self.properties = properties

    def add_property(self, property):
        self.properties[property.key] = property


class ServiceOfferProperty:
    def __init__(self, title, type, options):
        self.title = title
        self.type = type
        self.options = options


class ProgressResponse:
    def __init__(self, chain_links):
        self.chain_links = chain_links

    def add_chain_link(self, chain_link):
        self.chain_links[chain_link.key] = chain_link
        setattr(self, chain_link.key, chain_link)


class ProgressResponseChainLink:
    def __init__(self, message_id, service_catalogue_uuid, progress_updates):
        self.message_id = message_id
        self.service_catalogue_uuid = service_catalogue_uuid
        self.progress_updates = progress_updates


class ProgressUpdate:
    def __init__(
        self, standard_header, input_data, output_data, service_catalogue_uuid,
        permissible_service_provider_uuid, service_provider_header, service_offer_description,
        service_offer_parameters, service_offer_name, execution_parameters,
        offer_execution_request_to_cancel, execution_details, exec_start_timestamp,
        decline_reason_text, exec_decline_timestamp, exec_status_code, exec_status_text,
        exec_update_timestamp, exec_comp_text, exec_comp_code, results_dataset,
        exec_completion_timestamp, unfulfilled_service_uuid, exec_fail_text, exec_fail_code,
        notification_type, subject_type_name, subject_name, notification_message_url,
        notification_detail_url, notification_trigger_timestamp, publish_time, message_type,
        message_id, exec_url
    ):
        self.decline_reason_text = decline_reason_text
        self.exec_comp_code = exec_comp_code
        self.exec_comp_text = exec_comp_text
        self.exec_completion_timestamp = exec_completion_timestamp
        self.exec_decline_timestamp = exec_decline_timestamp
        self.exec_fail_code = exec_fail_code
        self.exec_fail_text = exec_fail_text
        self.exec_start_timestamp = exec_start_timestamp
        self.exec_status_code = exec_status_code
        self.exec_status_text = exec_status_text
        self.exec_update_timestamp = exec_update_timestamp
        self.exec_url = exec_url
        self.execution_details = execution_details
        self.execution_parameters = execution_parameters
        self.input_data = input_data
        self.message_id = message_id
        self.message_type = message_type
        self.notification_detail_url = notification_detail_url
        self.notification_message_url = notification_message_url
        self.notification_trigger_timestamp = notification_trigger_timestamp
        self.notification_type = notification_type
        self.offer_execution_request_to_cancel = offer_execution_request_to_cancel
        self.output_data = output_data
        self.permissible_service_provider_uuid = permissible_service_provider_uuid
        self.publish_time = publish_time
        self.results_dataset = results_dataset
        self.service_catalogue_uuid = service_catalogue_uuid
        self.service_offer_description = service_offer_description
        self.service_offer_name = service_offer_name
        self.service_offer_parameters = service_offer_parameters
        self.service_provider_header = service_provider_header
        self.standard_header = standard_header
        self.subject_name = subject_name
        self.subject_type_name = subject_type_name
        self.unfulfilled_service_uuid = unfulfilled_service_uuid


class StandardHeader:
    def __init__(
        self, app_message_uuid, request_message_uuid, originating_message_uuid, service_chain_uuid,
        message_timestamp, user_token, user_id, service_chain
    ):
        self.app_message_uuid = app_message_uuid
        self.request_message_uuid = request_message_uuid
        self.originating_message_uuid = originating_message_uuid
        self.service_chain_uuid = service_chain_uuid
        self.message_timestamp = message_timestamp
        self.user_token = user_token
        self.user_id = user_id
        self.service_chain = service_chain

class DatasetRequest:
    def __init__(self, data_uri, metadata):
        self.data_uri = data_uri
        self.metadata = metadata

class InputData:
    def __init__(self, data_token, dataset_uuid, dataset_type, cdf_metadata):
        self.data_token = data_token
        self.dataset_uuid = dataset_uuid
        self.dataset_type = dataset_type
        self.cdf_metadata = cdf_metadata


class OutputData:
    def __init__(self, file_uri, dataset_uuid, dataset_type):
        self.file_uri = file_uri
        self.dataset_uuid = dataset_uuid
        self.dataset_type = dataset_type


class ServiceProviderHeader:
    def __init__(
        self, service_provider_uuid, service_offer_uuid, service_request_uuid,
        service_provider_signature
    ):
        self.service_provider_uuid = service_provider_uuid
        self.service_offer_uuid = service_offer_uuid
        self.service_request_uuid = service_request_uuid
        self.service_provider_signature = service_provider_signature


class ExecutionDetails:
    def __init__(self, exec_request_uuid, exec_process_uuid):
        self.exec_request_uuid = exec_request_uuid
        self.exec_process_uuid = exec_process_uuid


class ResultsDataset:
    def __init__(self, file_uri, dataset_uuid, dataset_type):
        self.file_uri = file_uri
        self.dataset_uuid = dataset_uuid
        self.dataset_type = dataset_type


class ServiceChainParameters:
    def __init__(self, description_text):
    # def __init__(self, data_path, description_text):
        # self.data_path = data_path
        self.description_text = description_text


class ServiceChain:
    def __init__(
        self, service_chain_uuid, service_offer_uuid, service_provider_uuid,
        service_provider_signature, originating_message_uuid, service_request_uuid, parameters
    ):
        self.service_chain_uuid = service_chain_uuid
        self.service_offer_uuid = service_offer_uuid
        self.service_provider_uuid = service_provider_uuid
        self.service_provider_signature = service_provider_signature
        self.originating_message_uuid = originating_message_uuid
        self.service_request_uuid = service_request_uuid
        self.parameters = parameters


class InputDataset:
    def __init__(
        self, kind, id, self_link, media_link, name, bucket, generation, metageneration,
        content_type, storage_class, size, md5_hash, crc32c, etag, time_created, updated,
        time_storage_class_updated
    ):
        self.kind = kind
        self.id = id
        self.self_link = self_link
        self.media_link = media_link
        self.name = name
        self.bucket = bucket
        self.generation = generation
        self.metageneration = metageneration
        self.content_type = content_type
        self.storage_class = storage_class
        self.size = size
        self.md5_hash = md5_hash
        self.crc32c = crc32c
        self.etag = etag
        self.time_created = time_created
        self.updated = updated
        self.time_storage_class_updated = time_storage_class_updated


class InputDatasetsResponse:
    def __init__(self, kind, next_page_token, items):
        self.kind = kind
        self.next_page_token = next_page_token
        self.items = items


class ServiceRequestChainLink:
    def __init__(self, service, datasets):
        self.service = service
        self.datasets = datasets


class ServiceRequest:
    def __init__(self, service_chain):
        self.service_chain = service_chain

    def chain_link_count(self):
        return len(self.service_chain.keys())

    def add_chain_link(self, chain_link):
        index = self.chain_link_count() + 1
        self.service_chain[f'chain_link_{index}'] = chain_link


class ServiceExecutionRequestChainLinkParameters:
    def __init__(self, data_paths, description_text):
        self.data_paths = data_paths
        self.description_text = description_text

    def add_parameter(self, key, value):
        setattr(self, key, value)


class ServiceExecutionRequestChainLink:
    def __init__(
        self, service_chain_uuid, service_offer_uuid, service_provider_uuid,
        service_provider_signature, originating_message_uuid, service_request_uuid, parameters
    ):
        self.service_chain_uuid = service_chain_uuid
        self.service_offer_uuid = service_offer_uuid
        self.service_provider_uuid = service_provider_uuid
        self.service_provider_signature = service_provider_signature
        self.originating_message_uuid = originating_message_uuid
        self.service_request_uuid = service_request_uuid
        self.parameters = parameters


class ServiceExecutionRequest:
    def __init__(self, service_chain):
        self.service_chain = service_chain


class ServiceParameters:
    def __init__(self, params):
        self.params = params.properties.values()
        self.configuration = {}

    def describe(self):
        for parameter in self.params:
            try:
                print('PARAMETER_TITLE:', parameter.title)
            except:
                pass
            try:
                print('PARAMETER_TYPE:', parameter.type)
            except:
                pass
            try:
                print('PARAMETER_OPTIONS:', parameter.options)
            except:
                pass
            try:
                print('PARAMETER_DESCRIPTION:', parameter.description)
            except:
                pass

    def items(self):
        return self.params

    def configure(self, title, value):
        self.configuration[title] = value


class Service:
    def __init__(self, service_discovery_data, chain_link=''):
        self.service = service_discovery_data.service
        self.name = self.service.name
        self.params = ServiceParameters(self.service.parameters)
        self.service_discovery_data = service_discovery_data
        self.chain_link = chain_link


class WorkflowServiceChain:
    def __init__(self, datasets, services):
        self.datasets = datasets
        self.services = services
        self.chain_link_index = 1

    def add_service(self, service):
        service.chain_link = f'chain_link_{self.chain_link_index}'
        self.services.append(service)
        self.chain_link_index += 1

    def add_dataset(self, dataset, metadata=""):
        if dataset.dataset_type != 'GCS' and dataset.dataset_type != 'CDF':
            raise Exception(f'Unknown dataset type: "{dataset.dataset_type}"')
        if metadata != "":
            dataset_with_metadata = dataset
            dataset_with_metadata.metadata = metadata
            self.datasets.append(dataset_with_metadata)
        else:
            self.datasets.append(dataset)


class CommandInput:
    def __init__(self, response, service_chain):
        self.response = response
        self.service_chain = service_chain


class IngestionRequest:
    def __init__(self, action_type, target_system, data_urls):
        self.action_type = action_type
        self.target_system = target_system
        self.data_urls = data_urls


class DataURL:
    def __init__(self, data_url, data_type, seismic_metadata):
        self.data_url = data_url
        self.data_type = data_type
        self.seismic_metadata = seismic_metadata


class SeismicMetadata:
    def __init__(self, survey, file):
        self.survey = survey
        self.file = file


class Survey:
    def __init__(self, survey_name, survey_id):
        self.survey_name = survey_name
        self.survey_id = survey_id


class SeismicFile:
    def __init__(self, crs, custom_metadata):
        self.crs = crs
        self.custom_metadata = custom_metadata


class CustomMetadataItem:
    def __init__(self, metadata_key, metadata_value):
        self.metadata_key = metadata_key
        self.metadata_value = metadata_value


class DataID:
    def __init__(self, url, name, id, status):
        self.url = url
        self.name = name
        self.id = id
        self.status = status


class IngestionProgressResponse:
    def __init__(self, data_ids):
        self.data_ids = data_ids



from swap.api.clients import UBLClient
from swap.api.models import WorkflowServiceChain, CommandInput
from swap.api.sequences.builders import DefaultCommandSequenceBuilder


class Workflow:
    def __init__(self):
        self.service_chain_uuid = None
        self.service_chains = {}
        self.ubl_client = UBLClient()
        self.command_sequence = DefaultCommandSequenceBuilder().build()

    def create_service_chain(self, name):
        self.service_chains[name] = WorkflowServiceChain(datasets=[], services=[])

    def submit(self):
        output = {}

        for key, service_chain in self.service_chains.items():
            command_input = CommandInput(response=None, service_chain=service_chain)
            output[key] = self.command_sequence.call(command_input)

        return output

    def service_chain(self, name):
        return self.service_chains[name]



#!/usr/bin/env python
from jinja2 import Template
from os import path
from uuid import uuid4
import webbrowser
import requests
import json
from swap.common.config import Settings

config_template = """from swap.messaging.models import Config
from service.parameters import parameters

service_config = Config(
    project='{{ project }}',
    topic='{{ topic }}',
    subscription='{{ subscription }}',
    catalogue_uuid='{{ catalogue_uuid }}',
    dataset_type='{{ dataset_type }}',
    name='{{ name }}',
    service_description='{{ service_description }}',
    service_provider_name='{{ service_provider_name }}',
    service_provider_signature='{{ service_provider_signature }}',
    service_provider_uuid='{{ service_provider_uuid }}',
    icon_base64='{{ icon_base64 }}',
    icon_url='{{ icon_url }}',
    parameters=parameters
)"""


class ConfigGenerator:
    def __init__(self, config_output_path, parameters):
        self.template = Template('')
        self.config_output_path = config_output_path
        self.parameters = parameters
        self.swap_api_token= Settings.SWAP_API_TOKEN
        self.template = Template(config_template)

    def generate_subscription(self):
        # Call UBL to generate pub/sub subscription
        subscription_generation_url = 'https://ubl.stage.baringa-pharos.com/template_app/subscription'
        response = requests.get(subscription_generation_url,
                                params={
                                    "api_key": self.swap_api_token
                                })

        response_as_dict = json.loads(response.text)
        new_subscription_id = response_as_dict['subscription_id']

        return new_subscription_id

    def render(self, subscription):
        name = input('Enter a name for your service: ')
        name = name + '_dev'
        description = input('Enter a description for your service: ')

        return self.template.render(
            project='microservice-pilot',
            topic='akerbp-stage-pre-seismic-processing',
            subscription=subscription,
            catalogue_uuid=uuid4(),
            dataset_type='SEGY',
            name=name,
            service_description=description,
            service_provider_name='AkerBP',
            service_provider_signature=uuid4(),
            service_provider_uuid=uuid4(),
            icon_base64='ICON_BASE64',
            icon_url='https://ui.dev.baringa-pharos.com/static/media/Skua-salt-2010.4fa0f606.jpg',
            parameters='self.parameters'
        )

    def write(self, data):
        with open(self.config_output_path, 'w') as file:
            file.write(data)

    def generate(self):
        subscription = self.generate_subscription()
        data = self.render(subscription)
        self.write(data)

        print(f'A new service config has been successfully written to {path.abspath(self.config_output_path)}')



from swap.messaging.clients import MessagePublisherClient
from swap.messaging.messages.builders import (
    ExecutionFailureBuilder,
    ExecutionRequestAcceptBuilder,
    ExecutionStatusUpdateBuilder,
    ExecutionCompletionBuilder
)


class PubSubMessage:
    def __init__(self, config):
        self.message = ''
        self.client = MessagePublisherClient(config=config)

    def publish(self):
        self.client.publish(message=self.message.encode(), message_type=self.message_type)


class ExecutionFailureMessage(PubSubMessage):
    def __init__(
        self, message_id, config, offer_execution_request, execution_details={}, exec_fail_code='',
        exec_fail_text=''
    ):
        super().__init__(config=config)

        self.message_id = message_id
        self.message_type = 'ExecutionFailure'
        self.config = config
        self.offer_execution_request = offer_execution_request
        self.execution_details = execution_details
        self.exec_fail_code = exec_fail_code
        self.exec_fail_text = exec_fail_text

        self.message = self._build()

    def _build(self):
        builder = ExecutionFailureBuilder(
            message_id=self.message_id,
            offer_execution_request=self.offer_execution_request,
            execution_details=self.execution_details,
            exec_fail_code=self.exec_fail_code,
            exec_fail_text=self.exec_fail_text
        )

        return builder.build()


class ExecutionRequestAcceptMessage(PubSubMessage):
    def __init__(self, message_id, config, offer_execution_request, execution_details={}):
        super().__init__(config=config)

        self.message_id = message_id
        self.message_type = 'ExecutionRequestAccept'
        self.config = config
        self.offer_execution_request = offer_execution_request
        self.execution_details = execution_details

        self.message = self._build()

    def _build(self):
        builder = ExecutionRequestAcceptBuilder(
            message_id=self.message_id,
            offer_execution_request=self.offer_execution_request,
            execution_details=self.execution_details
        )

        return builder.build()


class ExecutionStatusUpdateMessage(PubSubMessage):
    def __init__(
        self, message_id, config, offer_execution_request, execution_details={},
        exec_status_code='', exec_status_text=''
    ):
        super().__init__(config=config)

        self.message_id = message_id
        self.message_type = 'ExecutionStatusUpdate'
        self.config = config
        self.offer_execution_request = offer_execution_request
        self.execution_details = execution_details
        self.exec_status_code = exec_status_code
        self.exec_status_text = exec_status_text

        self.message = self._build()

    def _build(self):
        builder = ExecutionStatusUpdateBuilder(
            message_id=self.message_id,
            offer_execution_request=self.offer_execution_request,
            execution_details=self.execution_details,
            exec_status_code=self.exec_status_code,
            exec_status_text=self.exec_status_text
        )

        return builder.build()


class ExecutionCompletionMessage(PubSubMessage):
    def __init__(
        self, message_id, config, offer_execution_request, execution_details={},
        output_data_id = '', exec_comp_code='', exec_comp_text=''
    ):
        super().__init__(config=config)

        self.message_id = message_id
        self.message_type = 'ExecutionCompletion'
        self.config = config
        self.offer_execution_request = offer_execution_request
        self.execution_details = execution_details
        self.output_data_id = output_data_id
        self.exec_comp_code = exec_comp_code
        self.exec_comp_text = exec_comp_text

        self.message = self._build()

    def _build(self):
        builder = ExecutionCompletionBuilder(
            message_id=self.message_id,
            offer_execution_request=self.offer_execution_request,
            execution_details=self.execution_details,
            output_data_id = self.output_data_id,
            exec_comp_code=self.exec_comp_code,
            exec_comp_text=self.exec_comp_text
        )

        return builder.build()



from swap.api.filters import (
    InputDatasetsResponseFilter, ServiceOffersResponseFilter, ProgressResponseFilter
)
from swap.api.schemas import (
    InputDatasetsResponseSchema, ServiceOffersResponseSchema, ProgressResponseSchema
)
from swap.common.parsers import Parser


class InputDatasetsResponseParser(Parser):
    def __init__(self):
        self.filter = InputDatasetsResponseFilter()
        self.schema = InputDatasetsResponseSchema()


class ServiceOffersResponseParser(Parser):
    def __init__(self):
        self.filter = ServiceOffersResponseFilter()
        self.schema = ServiceOffersResponseSchema()


class ProgressResponseParser(Parser):
    def __init__(self):
        self.filter = ProgressResponseFilter()
        self.schema = ProgressResponseSchema()



import base64
import json
import sys
from swap.common.logging import logger
from swap.messaging.clients import MessagePublisherClient
from swap.messaging.identities import OfferExecutionRequestIdentity, ServiceRequestIdentity
from swap.messaging.messages.builders import ServiceDiscoveryResponseBuilder, ServiceOfferingBuilder
from swap.messaging.parsers import MessageParser
from swap.messaging import ExecutionFailureMessage, ExecutionRequestAcceptMessage


class Action:
    def __init__(self, config):
        self.config = config
        self.message_parser = MessageParser()
        self.client = MessagePublisherClient(config=config)

    def parse_message(self, message):
        return self.message_parser.parse(message).data


class DefaultMessageAction(Action):
    def call(self, message_type, message, callback):
        logger.debug(f'Executing default message action for {message_type} message')


class IgnoredMessageAction(Action):
    def call(self, message_type, message, callback):
        logger.debug(f'Received a {message_type} message, ignoring')

        if message_type != 'ServiceDiscoveryResponse':
            loaded = json.loads(message.data)
            logger.debug(f'Message: {json.dumps(loaded, indent=2)}')


class ServiceDiscoveryAction(Action):
    def call(self, message_type, message, callback):
        logger.debug('Received a ServiceDiscovery message')

        builder = ServiceDiscoveryResponseBuilder(config=self.config)
        message = builder.build()

        logger.debug(f'Responding with: {message}')

        self.client.publish(message=message.encode(), message_type='ServiceDiscoveryResponse')


class ServiceRequestAction(Action):
    def __init__(self, config):
        super().__init__(config=config)
        self.identity = ServiceRequestIdentity()

    def respond_to(self, message_id, service_request):
        # FIXME Use when CDF integration finalised
        # decoded = base64.b64decode(service_request.input_data.cdf_metadata)
        # print('Base64 CDF metadata:', decoded)

        logger.info('Responding to ServiceRequest')
        builder = ServiceOfferingBuilder(message_id, self.config, service_request)
        message = builder.build()

        logger.debug(f'Responding to ServiceRequest with {message}')

        self.client.publish(message=message.encode(), message_type='ServiceOffering')

    def call(self, message_type, message, callback):
        logger.info('Received a ServiceRequest message')

        loaded = json.loads(message.data)
        logger.debug(f'Message: {json.dumps(loaded, indent=2)}')

        service_request = self.parse_message(message)

        if self.identity.match(service_request, self.config):
            message_id = message.message_id
            self.respond_to(message_id=message_id, service_request=service_request)


class ExecutionRequestCancelAction(Action):
    # Ignore for now
    def call(self, message_type, message, callback):
        logger.debug('Received an ExecutionRequestCancel message. Currently unable to cancel.')


class OfferExecutionRequestAction(Action):
    def __init__(self, config):
        super().__init__(config=config)
        self.identity = OfferExecutionRequestIdentity()

    def execution_parameters(self, offer_execution_request):
        execution_parameters_string = offer_execution_request.execution_parameters

        try:
            decoded = base64.b64decode(execution_parameters_string[2:-1])
            parsed = json.loads(decoded)
            return parsed
        except:
            e = sys.exc_info()
            logger.warn(e)
            return {}

    def respond_with_acceptance(self, message_id, offer_execution_request):
        message = ExecutionRequestAcceptMessage(
            message_id=message_id,
            config=self.config,
            offer_execution_request=offer_execution_request,
            execution_details='',
        )

        message.publish()

    def respond_with_failure(self, message_id, offer_execution_request, error):
        message = ExecutionFailureMessage(
            message_id=message_id,
            config=self.config,
            offer_execution_request=offer_execution_request,
            execution_details='',
            exec_fail_code='1',
            exec_fail_text='Execution failed'
        )

        message.publish()

    def respond_to(self, message_id, offer_execution_request, callback):
        self.respond_with_acceptance(message_id, offer_execution_request)

        parameters = self.execution_parameters(offer_execution_request)

        try:
            callback.call(
                message_id=message_id,
                offer_execution_request=offer_execution_request,
                parameters=parameters
            )
        except:
            e = sys.exc_info()
            logger.warn(f'Callback failed {e}')

            self.respond_with_failure(
                message_id=message_id,
                offer_execution_request=offer_execution_request,
                error=e
            )

    def call(self, message_type, message, callback):
        logger.info('Received an OfferExecutionRequest message')

        loaded = json.loads(message.data)
        logger.debug(f'Message: {json.dumps(loaded, indent=2)}')

        offer_execution_request = self.parse_message(message)

        if self.identity.match(offer_execution_request, self.config):
            message_id = message.message_id

            self.respond_to(
                message_id=message_id,
                offer_execution_request=offer_execution_request,
                callback=callback
            )



class ServiceDiscoveryVendor:
    def __init__(self, service_provider_name, service_provider_uuid, service_provider_signature):
        self.service_provider_name = service_provider_name
        self.service_provider_uuid = service_provider_uuid
        self.service_provider_signature = service_provider_signature


class ServiceDiscoveryProperty:
    def __init__(self, title, type, options=[], description=""):
        self.title = title
        self.type = type
        self.options = options
        self.description = description


class ServiceDiscoveryParameters:
    def __init__(self, properties):
        self.properties = properties


class ServiceDiscoveryService:
    def __init__(self, name, catalogue_uuid, service_description, parameters):
        self.name = name
        self.catalogue_uuid = catalogue_uuid
        self.service_description = service_description
        self.parameters = parameters


class ServiceDiscoveryDataset:
    def __init__(self, input, output):
        self.input = input
        self.output = output


class ServiceDiscoveryIcon:
    def __init__(self, base64, url):
        self.base64 = base64
        self.url = url


class ServiceDiscoveryData:
    def __init__(self, vendor, service, dataset, icon):
        self.vendor = vendor
        self.service = service
        self.dataset = dataset
        self.icon = icon


class ServiceDiscoveryResponse:
    def __init__(self, data):
        self.data = data



import jsonref
from jsonschema import validate
from os import path
from os import name as os_name


class Validator:
    def __init__(self):
        self.schema_path = path.join(path.dirname(__file__), f'static/schemas/{self.schema_name}')

        base_path = path.dirname(self.schema_path)
        base_uri = f'file://{base_path}/'

        if os_name == 'nt':
            base_uri = 'file:///{}/'.format(base_path)

        with open(self.schema_path) as schema:
            self.schema = jsonref.loads(schema.read(), base_uri=base_uri, jsonschema=True)

    def validate(self, instance):
        validate(instance=instance, schema=self.schema)
        return True


class ExecutionCompletionValidator(Validator):
    def __init__(self):
        self.schema_name = 'execution_completion.json'
        super().__init__()


class ExecutionFailureValidator(Validator):
    def __init__(self):
        self.schema_name = 'execution_failure.json'
        super().__init__()


class ExecutionRequestValidator(Validator):
    def __init__(self):
        self.schema_name = 'execution_request.json'
        super().__init__()


class OfferExecutionRequestValidator(Validator):
    def __init__(self):
        self.schema_name = 'offer_execution_request.json'
        super().__init__()


class ExecutionRequestAcceptValidator(Validator):
    def __init__(self):
        self.schema_name = 'execution_request_accept.json'
        super().__init__()


class ExecutionRequestCancelValidator(Validator):
    def __init__(self):
        self.schema_name = 'execution_request_cancel.json'
        super().__init__()


class ExecutionRequestDeclineValidator(Validator):
    def __init__(self):
        self.schema_name = 'execution_request_decline.json'
        super().__init__()


class ExecutionStatusUpdateValidator(Validator):
    def __init__(self):
        self.schema_name = 'execution_status_update.json'
        super().__init__()


class ServiceDiscoveryValidator(Validator):
    def __init__(self):
        self.schema_name = 'service_discovery.json'
        super().__init__()


class ServiceDiscoveryResponseValidator(Validator):
    def __init__(self):
        self.schema_name = 'service_discovery_response.json'
        super().__init__()


class ServiceOfferingValidator(Validator):
    def __init__(self):
        self.schema_name = 'service_offering.json'
        super().__init__()


class ServiceRequestValidator(Validator):
    def __init__(self):
        self.schema_name = 'service_request.json'
        super().__init__()



from swap.common.logging import logger
from swap.api.sequences.step_codes import StepCode

class CommandSequence:
    def __init__(self):
        self.current_index = 0
        self.index = {}
        self.steps = []

    def add_step(self, key, step):
        self.steps.append(step)
        self.index[key] = self.current_index
        self.current_index += 1

    def call(self, input):
        result = input

        for step in self.steps:
            result = step.call(result)

            if result == StepCode.NoServiceRequestsReceived:
                logger.info("No service requests received")
                return {"Result":"No service requests have been received, stopping workflow"}
            elif result == StepCode.NoServiceOffersReceived:
                logger.info("No service offers received")
                return {"Result":"No service offers have been received, stopping workflow"}

        return result


import requests
from google.cloud import pubsub_v1
from swap.common.config import Settings
from urllib.parse import urljoin


class MessagePublisherClient:
    def __init__(self, config, publisher=pubsub_v1.PublisherClient()):
        self.project = config.project
        self.topic = config.topic
        self.publisher = publisher
        self.topic_path = f'projects/{self.project}/topics/{self.topic}'

    def publish(self, message, message_type):
        self.publisher.publish(self.topic_path, message, message_type=message_type)


class MessageSubscriberClient:
    def __init__(
        self,
        project,
        topic,
        subscription,
        subscriber=pubsub_v1.SubscriberClient()
    ):
        self.project = project
        self.topic = topic
        self.subscription = subscription
        self.subscriber = subscriber

        self.topic_path = f'projects/{project}/topics/{topic}'
        self.subscription_path = f'projects/{project}/subscriptions/{subscription}'

    def subscribe(self, callback):
        return self.subscriber.subscribe(self.subscription_path, callback)


class DASClient:
    def __init__(self):
        self.base_url = Settings.DAS_URL

    def url(self, path):
        return urljoin(self.base_url, path)

    def get(self, path, params):
        joined_url = self.url(path)
        response = requests.get(joined_url, params=params)

        return response.text

    def redeem_token(self, token):
        params = {'dataToken': token, 'connectionDetails': 'True'}
        return self.get(path='/data', params=params)



from swap.common.logging import logger
from swap.api.sequences.step_codes import StepCode

class CommandStep:
    def __init__(self, command, pre_hook, post_hook):
        self.command = command
        self.pre_hook = pre_hook
        self.post_hook = post_hook

    def check_for_errors(self, output):
        if output == StepCode.NoServiceRequestsReceived:
            return False
        elif output == StepCode.NoServiceOffersReceived:
            return False
        else:
            return True

    def call(self, input):
        pre_hook_output = self.pre_hook.call(input)
        if self.check_for_errors(pre_hook_output) == False:
            return pre_hook_output
        command_output = self.command.call(pre_hook_output)
        if self.check_for_errors(command_output) == False:
            return command_output
        post_hook_output = self.post_hook.call(command_output)
        if self.check_for_errors(post_hook_output) == False:
            return post_hook_output

        return post_hook_output



from marshmallow import Schema, fields, post_load, EXCLUDE
from swap.api.models import (
    ExecutionDetails,
    InputData,
    InputDataset,
    InputDatasetsResponse,
    OutputData,
    ProgressResponse,
    ProgressResponseChainLink,
    ProgressUpdate,
    ResultsDataset,
    ServiceChain,
    ServiceChainParameters,
    ServiceOffer,
    ServiceOfferParameters,
    ServiceOfferProperty,
    ServiceOffersChainLink,
    ServiceOffersResponse,
    ServiceProviderHeader,
    ServiceRequestChainLink,
    StandardHeader,
    IngestionProgressResponse,
    DataID,
    DatasetRequest
)


class ServiceOfferPropertySchema(Schema):
    title = fields.Str()
    type = fields.Str()
    options = fields.List(fields.Str())

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceOfferProperty(**other)


class ServiceOfferParametersSchema(Schema):
    properties = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(ServiceOfferPropertySchema)
    )

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceOfferParameters(**other)


class ServiceOfferSchema(Schema):
    message_id = fields.Str()
    publish_time = fields.Str()
    parameters = fields.Nested(ServiceOfferParametersSchema)
    app_message_uuid = fields.Str()
    request_message_uuid = fields.Str()
    originating_message_uuid = fields.Str()
    service_chain_uuid = fields.Str()
    service_provider_uuid = fields.Str()
    service_offer_uuid = fields.Str()
    service_request_uuid = fields.Str()
    service_provider_signature = fields.Str()
    service_offer_name = fields.Str()
    service_offer_description = fields.Str()
    file_uris = fields.List(fields.Str())

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceOffer(**other)


class ServiceOffersChainLinkSchema(Schema):
    message_id = fields.Str()
    service_catalogue_uuid = fields.Str()
    service_offers = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(ServiceOfferSchema)
    )

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceOffersChainLink(**other)


class ServiceOffersResponseSchema(Schema):
    chain_links = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(ServiceOffersChainLinkSchema)
    )

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceOffersResponse(**other)


class ServiceChainParametersSchema(Schema):
    # data_path = fields.Str()
    description_text = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceChainParameters(**other)


class ServiceChainSchema(Schema):
    service_chain_uuid = fields.Str()
    service_offer_uuid = fields.Str()
    service_provider_uuid = fields.Str()
    service_provider_signature = fields.Str()
    originating_message_uuid = fields.Str()
    service_request_uuid = fields.Str()
    parameters = fields.Nested(ServiceChainParametersSchema)

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceChain(**other)


class StandardHeaderSchema(Schema):
    app_message_uuid = fields.Str()
    message_timestamp = fields.Str()
    originating_message_uuid = fields.Str()
    request_message_uuid = fields.Str()
    service_chain = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(ServiceChainSchema),
        allow_none=True
    )
    service_chain_uuid = fields.Str()
    user_id = fields.Str()
    user_token = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return StandardHeader(**other)


class DatasetRequestSchema(Schema):
    data_uri = fields.Str(allow_none=True)
    metadata = fields.Str(allow_none=True)

    @post_load
    def deserialise(self, other, **kwargs):
        return DatasetRequest(**other)


class InputDataSchema(Schema):
    cdf_metadata = fields.Str(allow_none=True)
    data_token = fields.Str(allow_none=True)
    dataset_type = fields.Str(allow_none=True)
    dataset_uuid = fields.Nested(DatasetRequestSchema)

    @post_load
    def deserialise(self, other, **kwargs):
        return InputData(**other)


class OutputDataSchema(Schema):
    dataset_type = fields.Str(allow_none=True)
    dataset_uuid = fields.Str(allow_none=True)
    file_uri = fields.Str(allow_none=True)

    @post_load
    def deserialise(self, other, **kwargs):
        return OutputData(**other)


class ServiceProviderHeaderSchema(Schema):
    service_offer_uuid = fields.Str()
    service_provider_signature = fields.Str()
    service_provider_uuid = fields.Str()
    service_request_uuid = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceProviderHeader(**other)


class ExecutionDetailsSchema(Schema):
    exec_request_uuid = fields.Str(allow_none=True)
    exec_process_uuid = fields.Str(allow_none=True)

    @post_load
    def deserialise(self, other, **kwargs):
        return ExecutionDetails(**other)


class ResultsDatasetSchema(Schema):
    file_uri = fields.Str(allow_none=True)
    dataset_uuid = fields.Str(allow_none=True)
    dataset_type = fields.Str(allow_none=True)

    @post_load
    def deserialise(self, other, **kwargs):
        return ResultsDataset(**other)


class ProgressUpdateSchema(Schema):
    decline_reason_text = fields.Str(allow_none=True)
    exec_comp_code = fields.Str(allow_none=True)
    exec_comp_text = fields.Str(allow_none=True)
    exec_completion_timestamp = fields.Str(allow_none=True)
    exec_decline_timestamp = fields.Str(allow_none=True)
    exec_fail_code = fields.Str(allow_none=True)
    exec_fail_text = fields.Str(allow_none=True)
    exec_start_timestamp = fields.Str(allow_none=True)
    exec_status_code = fields.Str(allow_none=True)
    exec_status_text = fields.Str(allow_none=True)
    exec_update_timestamp = fields.Str(allow_none=True)
    exec_url = fields.Str(allow_none=True)
    execution_details = fields.Nested(ExecutionDetailsSchema)
    execution_parameters = fields.Str(allow_none=True)
    input_data = fields.List(fields.Nested(InputDataSchema))
    message_id = fields.Str(allow_none=True)
    message_type = fields.Str(allow_none=True)
    notification_detail_url = fields.Str(allow_none=True)
    notification_message_url = fields.Str(allow_none=True)
    notification_trigger_timestamp = fields.Str(allow_none=True)
    notification_type = fields.Str(allow_none=True)
    offer_execution_request_to_cancel = fields.Str(allow_none=True)
    output_data = fields.Nested(OutputDataSchema, allow_none=True)
    permissible_service_provider_uuid = fields.Str(allow_none=True)
    publish_time = fields.Str(allow_none=True)
    results_dataset = fields.Nested(ResultsDatasetSchema, allow_none=True)
    service_catalogue_uuid = fields.Str(allow_none=True)
    service_offer_description = fields.Str(allow_none=True)
    service_offer_name = fields.Str(allow_none=True)
    service_offer_parameters = fields.Str(allow_none=True)
    service_provider_header = fields.Nested(ServiceProviderHeaderSchema)
    standard_header = fields.Nested(StandardHeaderSchema)
    subject_name = fields.Str(allow_none=True)
    subject_type_name = fields.Str(allow_none=True)
    unfulfilled_service_uuid = fields.Str(allow_none=True)

    @post_load
    def deserialise(self, other, **kwargs):
        return ProgressUpdate(**other)


class ProgressResponseChainLinkSchema(Schema):
    message_id = fields.Str()
    service_catalogue_uuid = fields.Str()
    progress_updates = fields.List(fields.Nested(ProgressUpdateSchema))

    @post_load
    def deserialise(self, other, **kwargs):
        return ProgressResponseChainLink(**other)


class ProgressResponseSchema(Schema):
    chain_links = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(ProgressResponseChainLinkSchema)
    )

    @post_load
    def deserialise(self, other, **kwargs):
        return ProgressResponse(**other)


class InputDatasetSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    kind = fields.Str()
    id = fields.Str()
    self_link = fields.Str()
    media_link = fields.Str()
    name = fields.Str()
    bucket = fields.Str()
    generation = fields.Str()
    metageneration = fields.Str()
    content_type = fields.Str()
    storage_class = fields.Str()
    size = fields.Str()
    md5_hash = fields.Str()
    crc32c = fields.Str()
    etag = fields.Str()
    time_created = fields.Str()
    updated = fields.Str()
    time_storage_class_updated = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return InputDataset(**other)


class InputDatasetsResponseSchema(Schema):
    kind = fields.Str()
    next_page_token = fields.Str()
    items = fields.List(fields.Nested(InputDatasetSchema))

    @post_load
    def deserialise(self, other, **kwargs):
        return InputDatasetsResponse(**other)


class ServiceRequestChainLinkSchema(Schema):
    service = fields.Str()
    datasets = fields.List(fields.Dict())

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceRequestChainLink(**other)


class ServiceRequestSchema(Schema):
    service_chain = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(ServiceRequestChainLinkSchema)
    )

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceRequestSchema(**other)


class ServiceExecutionRequestChainLinkParametersSchema(Schema):
    # data_path = fields.Str()
    description_text = fields.Str()


class ServiceExecutionRequestChainLinkSchema(Schema):
    service_chain_uuid = fields.Str()
    service_offer_uuid = fields.Str()
    service_provider_uuid = fields.Str()
    service_provider_signature = fields.Str()
    originating_message_uuid = fields.Str()
    service_request_uuid = fields.Str()
    parameters = fields.Nested(ServiceExecutionRequestChainLinkParametersSchema)


class ServiceExecutionRequestSchema(Schema):
    service_chain = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(ServiceExecutionRequestChainLinkSchema)
    )


class SurveySchema(Schema):
    survey_name = fields.Str()
    survey_id = fields.Str()


class CustomMetadataItemSchema(Schema):
    metadata_key = fields.Str()
    metadata_value = fields.Str()


class SeismicFileSchema(Schema):
    crs = fields.Str()
    custom_metadata = fields.List(fields.Nested(CustomMetadataItemSchema))


class SeismicMetadataSchema(Schema):
    survey = fields.Nested(SurveySchema)
    file = fields.Nested(SeismicFileSchema)


class DataURLSchema(Schema):
    data_url = fields.Str()
    data_type = fields.Str()
    seismic_metadata = fields.Nested(SeismicMetadataSchema)


class IngestionRequestSchema(Schema):
    action_type = fields.Str()
    target_system = fields.Str()
    data_urls = fields.List(fields.Nested(DataURLSchema))


class DataIDSchema(Schema):
    url = fields.Str()
    name = fields.Str()
    id = fields.Str()
    status = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return DataID(**other)


class IngestionProgressResponseSchema(Schema):
    data_ids = fields.List(fields.Nested(DataIDSchema))

    @post_load
    def deserialise(self, other, **kwargs):
        return IngestionProgressResponse(**other)



import logging
import sys
from swap.common.config import Settings

log_level = logging.getLevelName(Settings.LOG_LEVEL)

log_handler = logging.StreamHandler(sys.stdout)

logger = logging.getLogger()

logger.addHandler(log_handler)

logger.setLevel(log_level)



import re


class CaseConverter:
    def key_to_snake_case(self, key):
        return re.sub(r'(?<!^)(?=[A-Z])', '_', key).lower()

    def key_to_camel_case(self, key):
        split_key = key.split('_')
        output = []
        for index, item in enumerate(split_key):
            if index != 0:
                item = item.capitalize()

            output.append(item)

        return ''.join(output)

    def convert_to_camel_case(self, input):
        if type(input) == list:
            return [self.convert_to_camel_case(value) for value in input]

        if type(input) == dict:
            for key in input.keys():
                camel_case_key = self.key_to_camel_case(key)
                value = input.pop(key)

                if type(value) == dict or type(value) == list:
                    value = self.convert_to_camel_case(value)

                input[camel_case_key] = value

            return input

        return input

    def convert_to_snake_case(self, input):
        if type(input) == list:
            return [self.convert_to_snake_case(value) for value in input]

        if type(input) == dict:
            for key in input.keys():
                snake_case_key = self.key_to_snake_case(key)
                value = input.pop(key)

                if type(value) == dict or type(value) == list:
                    value = self.convert_to_snake_case(value)

                input[snake_case_key] = value

            return input

        return input



import json

class CDFDataset:
    def __init__(self, id, crs, survey_id, custom_metadata):
        self.id = id
        self.crs = crs
        self.survey_id = survey_id
        self.custom_metadata = custom_metadata

    def __str__(self):
        raw = {
            'id': self.id,
            'crs': self.crs,
            'survey_id': self.survey_id,
            'custom_metadata': self.custom_metadata
        }

        return json.dumps(raw, indent=2)


class GCSDataset:
    def __init__(self, name, url, size, custom_metadata):
        self.name = name
        self.url = url
        self.size = size
        self.custom_metadata = custom_metadata

    def __str__(self):
        raw = {
            'name': self.name,
            'url': self.url,
            'size': self.size,
            'custom_metadata': self.custom_metadata
        }

        return json.dumps(raw, indent=2)



import json
import sys
from jsonschema import ValidationError
from swap.common.logging import logger
from swap.common.schemas import ServiceDiscoveryResponseSchema, ServiceDiscoverySchema
from swap.common.validators import ServiceDiscoveryResponseValidator, ServiceDiscoveryValidator
from swap.messaging.models import Message
from swap.messaging.schemas import (
    ExecutionCompletionSchema,
    ExecutionFailureSchema,
    ExecutionRequestAcceptSchema,
    ExecutionRequestCancelSchema,
    ExecutionRequestDeclineSchema,
    ExecutionRequestSchema,
    ExecutionStatusUpdateSchema,
    OfferExecutionRequestSchema,
    ServiceOfferingSchema,
    ServiceRequestSchema
)
from swap.common.validators import (
    ExecutionCompletionValidator,
    ExecutionFailureValidator,
    ExecutionRequestAcceptValidator,
    ExecutionRequestCancelValidator,
    ExecutionRequestDeclineValidator,
    ExecutionRequestValidator,
    ExecutionStatusUpdateValidator,
    OfferExecutionRequestValidator,
    ServiceOfferingValidator,
    ServiceRequestValidator
)


class MessageParser:
    def _find_schema(self, message_type):
        mapping = {
            'ExecutionCompletion': ExecutionCompletionSchema(),
            'ExecutionFailure': ExecutionFailureSchema(),
            'ExecutionRequest': ExecutionRequestSchema(),
            'ExecutionRequestAccept': ExecutionRequestAcceptSchema(),
            'ExecutionRequestCancel': ExecutionRequestCancelSchema(),
            'ExecutionRequestDecline': ExecutionRequestDeclineSchema(),
            'ExecutionStatusUpdate': ExecutionStatusUpdateSchema(),
            'OfferExecutionRequest': OfferExecutionRequestSchema(),
            'ServiceDiscovery': ServiceDiscoverySchema(),
            'ServiceDiscoveryResponse': ServiceDiscoveryResponseSchema(),
            'ServiceOffering': ServiceOfferingSchema(),
            'ServiceRequest': ServiceRequestSchema()
        }

        return mapping.get(message_type)

    def _find_validator(self, message_type):
        mapping = {
            'ServiceDiscoveryResponse': ServiceDiscoveryResponseValidator(),
            'ExecutionCompletion': ExecutionCompletionValidator(),
            'ExecutionFailure': ExecutionFailureValidator(),
            'ExecutionRequest': ExecutionRequestValidator(),
            'ExecutionRequestAccept': ExecutionRequestAcceptValidator(),
            'ExecutionRequestCancel': ExecutionRequestCancelValidator(),
            'ExecutionRequestDecline': ExecutionRequestDeclineValidator(),
            'ExecutionStatusUpdate': ExecutionStatusUpdateValidator(),
            'OfferExecutionRequest': OfferExecutionRequestValidator(),
            'ServiceDiscovery': ServiceDiscoveryValidator(),
            'ServiceOffering': ServiceOfferingValidator(),
            'ServiceRequest': ServiceRequestValidator()
        }

        return mapping.get(message_type)

    def _validate(self, message_type, message_dict):
        validator = self._find_validator(message_type=message_type)
        validator.validate(instance=message_dict)

    def _deserialise(self, message_type, message_raw):
        message_dict = json.loads(message_raw)

        try:
            self._validate(message_type=message_type, message_dict=message_dict)
        except ValidationError as e:
            logger.info(f'Received invalid {message_type} message')
            raise Exception(e)

        schema = self._find_schema(message_type=message_type)

        return schema.load(message_dict)

    def parse(self, message):
        message_raw = message.data.decode('utf-8', 'ignore')
        message_type = message.attributes.get('message_type')

        message_data = None

        try:
            message_data = self._deserialise(message_type=message_type, message_raw=message_raw)
        except:
            e = sys.exc_info()
            logger.info(f'Could not deserialise {message_type}')
            logger.info(e)

        return Message(
            message_data=message_data,
            message_raw=message_raw,
            message_type=message_type
        )



from swap.api.clients import UBLClient
from swap.api.parsers import ProgressResponseParser
from swap.common.logging import logger
from time import sleep


class State:
    def get_keys(self, input):
        return [update.message_type for update in input.progress_updates]


class FinishedState(State):
    def __init__(self):
        self.name = 'finished'

    def run(self, input):
        logger.info('Entered finished state')

    def next(self, input):
        return input


class RunningState(State):
    def __init__(self):
        self.name = 'running'

    def run(self, input):
        logger.info('Entered running state')

    def next(self, input):
        keys = self.get_keys(input)

        if 'ExecutionCompletion' in keys:
            return FinishedState()

        if 'ExecutionFailure' in keys:
            return FinishedState()

        if 'ExecutionStatusUpdate' in keys:
            return RunningState()

        if 'ExecutionRequestAccept' in keys:
            return RunningState()

        raise Exception(f'No expected message types in {keys} for running state')


class WaitingState(State):
    def __init__(self):
        self.name = 'waiting'

    def run(self, input):
        logger.info('Entered waiting state')

        updates = input.progress_updates
        status_updates = [i for i in updates if i.message_type == 'ExecutionStatusUpdate']

        for update in status_updates:
            if update.exec_url != 'None':
                logger.info(f'Exec URL: {update.exec_url}')

    def next(self, input):
        keys = self.get_keys(input)

        if 'ExecutionCompletion' in keys:
            return FinishedState()

        if 'ExecutionFailure' in keys:
            return FinishedState()

        if 'ExecutionStatusUpdate' in keys:
            return RunningState()


class UnstartedState(State):
    def __init__(self):
        self.name = 'unstarted'

    def run(self, input):
        logger.info('Entered unstarted state')

    def next(self, input):
        keys = self.get_keys(input)

        if 'ExecutionRequestAccept' in keys:
            return RunningState()

        if 'ExecutionFailure' in keys:
            return FinishedState()

        raise Exception(f'No expected message types in {keys} in unstarted state')


class ProgressStateMachine:
    def __init__(self, service_chain_uuid, chain_link):
        self.chain_link = chain_link
        self.current_state = UnstartedState()
        self.data = {}
        self.parser = ProgressResponseParser()
        self.service_chain_uuid = service_chain_uuid
        self.ubl_client = UBLClient()
        self.sleep_time = 0

    def submit(self):
        url = f'/progress?service_chain_uuid={self.service_chain_uuid}'
        response = self.ubl_client.get(url)

        parsed_response = self.parser.parse(response)

        if parsed_response == 'JSONDecodeError':
            logger.info(f'The service chain uuid associated with this workflow is: {self.service_chain_uuid}')
            exit('')

        return parsed_response.chain_links[self.chain_link]

    def parse_status_updates(self, updates):
        status_updates = [i for i in updates if i.message_type == 'ExecutionStatusUpdate']

        self.data['update_messages'] = []

        for update in status_updates:
            if update.exec_status_text != 'None':
                self.data['update_messages'].append(update.exec_status_text)
            else:
                self.data['update_messages'].append('N/A')

            if update.exec_url != 'None':
                self.data['exec_url'] = update.exec_url

        self.data['update_messages'].reverse()

    def parse_execution_completion(self, updates):
        criteria = (update for update in updates if update.message_type == 'ExecutionCompletion')
        completion_update = next(criteria, None)

        if completion_update:
            self.data['status'] = 'complete'
            self.data['exec_comp_text'] = completion_update.exec_comp_text
            self.data['results_dataset'] = completion_update.results_dataset.file_uri

    def parse_execution_failure(self, updates):
        criteria = (update for update in updates if update.message_type == 'ExecutionFailure')

        failure_update = next(criteria, None)

        if failure_update:
            self.data['status'] = 'failed'
            self.data['exec_fail_text'] = failure_update.exec_fail_text

    def log_current_state(self, input):
        keys = self.current_state.get_keys(input)

        if 'ExecutionCompletion' in keys:
            logger.info('ExecutionCompletion')
        elif 'ExecutionFailure' in keys:
            logger.info('ExecutionFailure')
        elif 'ExecutionStatusUpdate' in keys:
            logger.info('ExecutionStatusUpdate')
            # FIXME KeyError
            try:
                logger.info(self.data['update_messages'][-1])
            except:
                pass
        elif 'ExecutionRequestAccept' in keys:
            logger.info('ExecutionRequestAccept')

    def update(self, response):
        updates = response.progress_updates
        self.parse_execution_completion(updates)
        self.parse_execution_failure(updates)
        self.parse_status_updates(updates)

    def run(self):
        while True:
            response = self.submit()
            self.current_state.run(response)
            self.update(response)
            self.log_current_state(response)

            if self.current_state.name == 'finished':
                break

            self.current_state = self.current_state.next(response)

            sleep(self.sleep_time)

        self.update(response)

        return self.data



import os
import warnings
import requests
import json

def check_environment_variables():
    try:
        key_exists = os.path.exists(os.getcwd() + '/auth/service_account_key.json')
    except:
        raise Exception("You need a Google Cloud service account key to use the SWAP sample app, and this should be saved in the 'auth' folder with the name 'service_account_key.json'")

    if "SWAP_API_TOKEN" not in os.environ:
        raise Exception("Your SWAP_API_TOKEN environment hasn't been set, this is required for using a SWAP service")

    if "COGNITE_API_KEY" not in os.environ:
        exception_message = "The environment variable COGNITE_API_KEY needs to be set before using the sample app."
        exception_message += "\n On a Windows system you can do this with the following command:"
        exception_message += "\n    set COGNITE_API_KEY=REPLACE_WITH_YOUR_COGNITE_API_KEY"
        exception_message += "\n On a Unix based system you can do this with the following command:"
        exception_message += "\n    export COGNITE_API_KEY=REPLACE_WITH_YOUR_COGNITE_API_KEY"
        print(exception_message)

def check_latest_version_of_sdk():
    print("Checking you have the latest version of the SDK... ", end = '')
    remote_sdk_info = requests.get("https://pypi.python.org/pypi/swap-python-sdk/json")
    remote_sdk_info_json = json.loads(remote_sdk_info.text)
    remote_sdk_version = remote_sdk_info_json['info']['version']

    from subprocess import call
    f = open("sdk_version.txt", "w+")
    if os.name == 'nt':
        call('pipenv run pip show swap-python-sdk | findstr Version', shell=True, stdout=f)
    else:
        call('pipenv run pip show swap-python-sdk | grep Version', shell=True, stdout=f)
    f.close()
    f = open("sdk_version.txt", "r")
    local_sdk_version = f.read().replace('\n','').replace('Version: ','')
    f.close()

    if remote_sdk_version == local_sdk_version:
        print("Your sdk version is up to date")
    else:
        print("\nIt looks like you are using an older version of the SDK")
        print(f"  Your version: {local_sdk_version}")
        print(f"  Latest pypi version: {remote_sdk_version}")
        print("To update the sdk, please run the following command: ")
        print("pipenv run pip install --upgrade swap-python-sdk")



from swap.api.models import (
    ServiceExecutionRequest,
    ServiceExecutionRequestChainLink,
    ServiceExecutionRequestChainLinkParameters,
    IngestionRequest,
    DataURL,
    SeismicMetadata,
    Survey,
    SeismicFile,
    CustomMetadataItem,
)
from swap.api.schemas import IngestionRequestSchema
from swap.common.utils import CaseConverter


class ServiceExecutionRequestBuilder:
    def build(self, input):
        service_chain = {}
        data_paths = []

        for dataset in input.service_chain.datasets:
            data_paths.append(dataset.name)

        for chain_link_key, service_offer in input.response.items():
            chain_link_parameters = ServiceExecutionRequestChainLinkParameters(
                data_paths=data_paths,
                description_text=''
            )

            service_execution_request_chain_link = ServiceExecutionRequestChainLink(
                service_chain_uuid=service_offer.service_chain_uuid,
                service_offer_uuid=service_offer.service_offer_uuid,
                service_provider_uuid=service_offer.service_provider_uuid,
                service_provider_signature=service_offer.service_provider_signature,
                originating_message_uuid=service_offer.originating_message_uuid,
                service_request_uuid=service_offer.service_request_uuid,
                parameters=chain_link_parameters
            )

            service_chain[chain_link_key] = service_execution_request_chain_link

        service_execution_request = ServiceExecutionRequest(
            service_chain=service_chain
        )

        return service_execution_request


class TriggerIngestionRequestBuilder:
    def build_custom_metadata(self):
        return [CustomMetadataItem(**item) for item in self.custom_metadata]

    def build_file(self):
        return SeismicFile(
            crs=self.crs,
            custom_metadata=self.build_custom_metadata()
        )

    def build_survey(self):
        return Survey(
            survey_id=self.survey_id,
            survey_name=''
        )

    def build_seismic_metadata(self):
        return SeismicMetadata(
            survey=self.build_survey(),
            file=self.build_file()
        )

    def build_data_url(self):
        return DataURL(
            data_url=self.data_url,
            data_type='SEISMIC',
            seismic_metadata=self.build_seismic_metadata()
        )

    def build(self, data_url, survey_id, crs, custom_metadata=[]):
        self.data_url = data_url
        self.survey_id = survey_id
        self.crs = crs
        self.custom_metadata = custom_metadata

        ingestion_request = IngestionRequest(
            action_type='TRIGGER_INGESTION',
            target_system='CDF',
            data_urls=[self.build_data_url()]
        )

        schema = IngestionRequestSchema()
        result = dict(schema.dump(ingestion_request))

        converter = CaseConverter()
        return converter.convert_to_camel_case(result)



from swap.messaging.listeners import MessageListener
from swap.messaging import (
    ExecutionStatusUpdateMessage, ExecutionCompletionMessage, ExecutionFailureMessage
)
from swap.messaging.clients import DASClient
import json
from swap.service.datasets import GCSDataset, CDFDataset

class SwapService:

    def __init__(self, service_config):
        self.service_config = service_config

    def send_execution_status_update(self, details='', code='0', text='Execution update'):
        print('Sending execution status update')
        execution_status_update = ExecutionStatusUpdateMessage(
            message_id=self.message_id,
            config=self.service_config,
            offer_execution_request=self.offer_execution_request,
            execution_details=details,
            exec_status_code=code,
            exec_status_text=text
        )

        execution_status_update.publish()

    def send_execution_completion(self, code='0', text='Execution complete', output_id='placeholder_id'):
        print('Sending execution completion')

        if output_id is not 'placeholder_id':
            self.set_output_id(output_id)

        execution_completion = ExecutionCompletionMessage(
            message_id=self.message_id,
            config=self.service_config,
            offer_execution_request=self.offer_execution_request,
            output_data_id = self.output_file_uri,
            exec_comp_code=code,
            exec_comp_text=text
        )

        execution_completion.publish()

    def send_execution_failure(self, code='0', text='Execution failed'):
        print('Sending execution failure update')

        execution_failure_message = ExecutionFailureMessage(
            message_id=self.message_id,
            config=self.service_config,
            offer_execution_request=self.offer_execution_request,
            exec_fail_code=code,
            exec_fail_text=text
        )

        execution_failure_message.publish()

    def _load_parsed_tokens(self):
        parsed_tokens = []
        for dataset in self.offer_execution_request.input_data:
            token = dataset.data_token
            client = DASClient()
            redeemed_token = client.redeem_token(token=token)
            parsed_token = json.loads(redeemed_token)
            parsed_tokens.append(parsed_token['links'][0])
        return parsed_tokens

    def _load_storage_types(self):
        storage_types = []
        for parsed_token in self.parsed_tokens:
            storage_types.append(parsed_token['connectionDetails']['storageType'])
        return storage_types

    def _load_gcp_data_parameters(self, parsed_token):
        name = parsed_token['metadata']['dataName']
        url = parsed_token['connectionDetails']['dataUrl']
        size = parsed_token['metadata']['dataSize']

        custom_metadata = ""
        try:
            metadata_ready_for_json = parsed_token['metadata']['customMetadata'].replace("'", '"')
            custom_metadata = json.loads(metadata_ready_for_json)
        except:
            print(f'Service tried to load custom metadata provided by user into dict but failed. Custom metadata:{parsed_token["metadata"]["customMetadata"]}')
            pass

        return GCSDataset(name=name, url=url, size=size, custom_metadata=custom_metadata)

    def _load_cdf_data_parameters(self, parsed_token):
        id = parsed_token['connectionDetails']['storageId']
        crs = ''
        survey_id = ''

        try:
            for item in parsed_token['metadata']['customMetadata']:
                if item['metadataKey'] == 'crs':
                    crs = item['metadataValue']

                if item['metadataKey'] == 'survey_id':
                    survey_id = item['metadataValue']
        except:
            pass

        custom_metadata = ""
        try:
            metadata_ready_for_json = parsed_token['metadata']['customMetadata'].replace("'", '"')
            custom_metadata = json.loads(metadata_ready_for_json)
        except:
            print(f'Service tried to load custom metadata provided by user into dict but failed. Custom metadata:{parsed_token["metadata"]["customMetadata"]}')
            pass

        return CDFDataset(id=id, crs=crs, survey_id=survey_id, custom_metadata=custom_metadata)

    def _load_data_parameters(self):
        self.data_details = []

        for parsed_token in self.parsed_tokens:
            if parsed_token['connectionDetails']['storageType'] == 'GCP':
                self.data_details.append(self._load_gcp_data_parameters(parsed_token))
            if parsed_token['connectionDetails']['storageType'] == 'CDF':
                self.data_details.append(self._load_cdf_data_parameters(parsed_token))

    def _load_output_file_uri(self):
        return self.offer_execution_request.output_data.file_uri

    def set_output_id(self, output_id):
        self.output_file_uri = output_id

    def start(self):
        print(f'Started service {self.service_config.name}')

        listener = MessageListener(config=self.service_config)
        listener.handler.register_callback(message_type='OfferExecutionRequest', callback=self)
        listener.listen()

    def call(self, message_id, offer_execution_request, parameters):

        if parameters is list:
            for unwanted_param in ['description_text']:
                del parameters[unwanted_param]

        self.message_id = message_id
        self.offer_execution_request = offer_execution_request
        self.parameters = parameters

        self.parsed_tokens = self._load_parsed_tokens()

        self.storage_types = self._load_storage_types()

        self.output_file_uri = self._load_output_file_uri()

        self._load_data_parameters()

        self.execute()



import re
from swap.common.filters import Filter
from swap.common.logging import logger
from swap.api.sequences.step_codes import StepCode

class InputDatasetsResponseFilter(Filter):
    def required_keys(self):
        return [
            'kind', 'id', 'self_link', 'media_link', 'name', 'bucket', 'generation',
            'metageneration', 'content_type', 'storage_class', 'size', 'md5_hash', 'crc32c', 'etag',
            'time_created', 'updated', 'time_storage_class_updated'
        ]

    def keys_to_correct(self):
        return [
            'selfLink', 'mediaLink', 'contentType', 'storageClass', 'md5Hash', 'timeCreated',
            'timeStorageClassUpdated'
        ]

    def correct_next_page_token(self):
        if 'nextPageToken' in self.response:
            next_page_token = self.response.pop('nextPageToken')
            self.response['next_page_token'] = next_page_token

    def correct_case(self):
        for index, item in enumerate(self.response['items']):
            for key in self.required_keys():
                if key not in item:
                    item[key] = ''

            for key in self.keys_to_correct():
                if key not in item:
                    continue

                value = item.pop(key)
                split_key = re.split('(?=[A-Z])', key)
                snake_case_key = '_'.join([x.lower() for x in split_key])

                item[snake_case_key] = value

            self.response['items'][index] = item

    def apply(self, response):
        self.response = response

        self.correct_next_page_token()
        self.correct_case()

        return self.response


class ServiceOffersResponseFilter(Filter):
    def wrap_response(self):
        self.response = {'chain_links': self.response}

    def correct_case(self):
        for chain_link_key, chain_link in self.response.items():
            if not chain_link['service_offers'].items():
                logger.info(f'No service offers received yet for service: {chain_link["service_catalogue_uuid"]}')
                return StepCode.NoServiceOffersReceived
            for service_offer_key, service_offer in chain_link['service_offers'].items():
                if 'messageId' in service_offer:
                    message_id_value = service_offer.pop('messageId')
                    service_offer['message_id'] = message_id_value

                if 'publishTime' in service_offer:
                    publish_time_value = service_offer.pop('publishTime')
                    service_offer['publish_time'] = publish_time_value

                if 'Parameters' in service_offer:
                    del service_offer['Parameters']
                    service_offer['parameters'] = {'properties': {}}

    def apply(self, response):
        self.response = response

        self.correct_case()
        self.wrap_response()

        return self.response


class ProgressResponseFilter(Filter):
    def correct_case_and_type(self):
        for chain_link_key, chain_link in self.response.items():
            for progress_update in chain_link['progress_updates']:
                message_id = progress_update.pop('messageId')
                progress_update['message_id'] = message_id

                publish_time = progress_update.pop('publishTime')
                progress_update['publish_time'] = publish_time

                if type(progress_update['standard_header']['service_chain']) == str:
                    progress_update['standard_header']['service_chain'] = {}

                if type(progress_update['input_data']) == str:
                    progress_update['input_data'] = {
                        'cdf_metadata': 'None',
                        'data_token': 'None',
                        'dataset_type': 'None',
                        'dataset_uuid': 'None'
                    }

                if type(progress_update['output_data']) == str:
                    progress_update['output_data'] = {
                        'dataset_type': 'None',
                        'dataset_uuid': 'None',
                        'file_uri': 'None'
                    }

                if type(progress_update['results_dataset']) == str:
                    progress_update['results_dataset'] = {
                        'file_uri': 'None',
                        'dataset_uuid': 'None',
                        'dataset_type': 'None'
                    }

    def apply(self, response):
        self.response = response

        self.correct_case_and_type()

        return {'chain_links': self.response}
